/*++
 Copyright (c) 2008, Intel Corporation
 All rights reserved. 

 This UEFI SCT ("Software") is furnished under license and may only be used or copied
 in accordance with the terms of that license UEFI Contributor Agreement.
 No license, express or implied, by estoppel or otherwise, to any intellectual property
 rights is granted by this document. The Software is subject to change without notice,
 and should not be construed as a commitment by Intel Corporation to market, license,
 sell or support any product or technology. Unless otherwise provided for in the license
 under which this Software is provided, the Software is provided AS IS, with no warranties
 of any kind, express or implied. Except as expressly permitted by the Software license,
 neither Intel Corporation nor its suppliers assumes any responsibility or liability
 for any errors or inaccuracies that may appear herein.
--*/
/*++

Module Name:
  
    EmsProtoTcp.c
    
Abstract:

    Implementation for functions for protocol TCP

--*/

#include "EmsNet.h"
#include "EmsTclInit.h"
#include "EmsPktMain.h"
#include "EmsPktCreate.h"
#include "EmsProtocols.h"
#include "EmsProtoEth.h"
#include "EmsProtoIp.h"
#include "EmsProtoTcp.h"
#include "EmsUtilityString.h"

STATIC
libnet_t          *
TcpCreatePacket (
  IN INT32 Argc,
  IN INT8  *Argv[]
  );

STATIC
FIELD_T           *
TcpUnpackPacket (
  IN UINT8  *Packet,
  IN UINT32 Length
  );

PROTOCOL_ENTRY_T  TcpProtocol = {
  "TCP",            /* name */
  TcpCreatePacket,  /* Create packet routine */
  TcpUnpackPacket,  /* Unpacket routine */
  NULL              /* TCP TCL initialzation routine*/
};

STATIC UINT16     TcpSrcPort;
STATIC UINT16     TcpDstPort;
STATIC UINT32     TcpSeq;
STATIC UINT32     TcpAck;
STATIC UINT8      TcpDataOffset;
STATIC UINT8      TcpCtrlFlag;
STATIC UINT16     TcpWinSize;
STATIC UINT16     TcpChecksum;
STATIC UINT16     TcpUrgent;
STATIC PAYLOAD_T  TcpOptions  = { NULL, 0 };
STATIC PAYLOAD_T  TcpPayload  = { NULL, 0 };

#define DEFAULT_SEQ_NUM   0x00
#define DEFAULT_ACK_NUM   0x00
#define DEFAULT_SRC_PORT  0x00
#define DEFAULT_CTRL_FLAG 0x00
#define DEFAULT_WIN_SIZE  0x400
#define DEFAULT_CHECKSUM  0x00
#define DEFAULT_URGENT    0x00

STATIC BOOLEAN    SrcPortOk;
STATIC BOOLEAN    DstPortOk;
STATIC BOOLEAN    SeqOk;
STATIC BOOLEAN    AckOk;
STATIC BOOLEAN    DataOffsetOk;
STATIC BOOLEAN    CtrlFlagOk;
STATIC BOOLEAN    WinSizeOk;
STATIC BOOLEAN    ChecksumOk;
STATIC BOOLEAN    UrgentOk;
STATIC BOOLEAN    OptionsOk;
STATIC BOOLEAN    PayloadOk;

FIELD_T           TcpField[] = {
  /* name         type         value           Mandatory Isset?*/
  {
    "Tcp_sp",
    OCTET2,
    &TcpSrcPort,
    FALSE,
    &SrcPortOk
  },
  {
    "Tcp_dp",
    OCTET2,
    &TcpDstPort,
    TRUE,
    &DstPortOk
  },
  {
    "Tcp_seq",
    OCTET4,
    &TcpSeq,
    FALSE,
    &SeqOk
  },
  {
    "Tcp_ack",
    OCTET4,
    &TcpAck,
    FALSE,
    &AckOk
  },
  {
    "Tcp_offset",
    OCTET1,
    &TcpDataOffset,
    FALSE,
    &DataOffsetOk
  },
  {
    "Tcp_control",
    OCTET1,
    &TcpCtrlFlag,
    FALSE,
    &CtrlFlagOk
  },
  {
    "Tcp_win",
    OCTET2,
    &TcpWinSize,
    FALSE,
    &WinSizeOk
  },
  {
    "Tcp_sum",
    OCTET2,
    &TcpChecksum,
    FALSE,
    &ChecksumOk
  },
  {
    "Tcp_urg",
    OCTET2,
    &TcpUrgent,
    FALSE,
    &UrgentOk
  },
  {
    "Tcp_options",
    PAYLOAD,
    &TcpOptions,
    FALSE,
    &OptionsOk
  },
  {
    "Tcp_payload",
    PAYLOAD,
    &TcpPayload,
    FALSE,
    &PayloadOk
  },
  {
    NULL
  }
};

STATIC
libnet_t *
TcpCreatePacket (
  INT32 Argc,
  INT8  *Argv[]
  )
/*++

Routine Description:

  Create a TCP packet

Arguments:

  Argc        - Argument counter.
  Argv        - Argument value pointer array.

Returns:

  NULL if error occurs
  The pointer to a libnet_t type's object if successfully

--*/
{
  libnet_t      *l;
  libnet_ptag_t Tag;
  UINT32        Index;
  INT8          ErrBuf[LIBNET_ERRBUF_SIZE];
  UINT32        Len;
  EMS_STATUS    Status;

  //
  // Initialize all the local variables
  //
  for (Index = 0; TcpField[Index].Name; Index++) {
    *TcpField[Index].IsOk = FALSE;
    Status                = GetValueByName (&TcpField[Index], Argc, Argv);
    if (IsError (Status)) {
      return NULL;
    }
  }
  //
  //  Initialize the libnet
  //
  l = libnet_init (
        LIBNET_LINK,
        /* injection type */EmsInterface,
        /* network interface */ErrBuf /* errbuf */
        );
  if (NULL == l) {
    return NULL;
  }

  /*
   * Build TCP options
   */
  Tag = libnet_build_tcp_options (
          OptionsOk ? TcpOptions.Payload : NULL,
          OptionsOk ? TcpOptions.Len : 0,
          l,
          0
          );

  if (-1 == Tag) {
    return NULL;
  }

  /*
  * Build TCP packet
  */
  Len = LIBNET_TCP_H;
  Len += PayloadOk ? TcpPayload.Len : 0;

  Tag = libnet_build_tcp (
          SrcPortOk ? TcpSrcPort : DEFAULT_SRC_PORT,
          TcpDstPort,
          SeqOk ? TcpSeq : DEFAULT_SEQ_NUM,
          AckOk ? TcpAck : DEFAULT_ACK_NUM,
          CtrlFlagOk ? TcpCtrlFlag : DEFAULT_CTRL_FLAG,
          WinSizeOk ? TcpWinSize : DEFAULT_WIN_SIZE,
          ChecksumOk ? TcpChecksum : DEFAULT_CHECKSUM,
          TcpUrgent ? TcpUrgent : DEFAULT_URGENT,
          Len,
          PayloadOk ? TcpPayload.Payload : NULL,
          PayloadOk ? TcpPayload.Len : 0,
          l,
          0
          );

  if (-1 == Tag) {
    return NULL;
  }

  Len = LIBNET_IPV4_H + LIBNET_TCP_H;
  Len += PayloadOk ? TcpPayload.Len : 0;

  Tag = libnet_build_ipv4 (
          Len,
          0,
          /* TOS */GetIpId (),
          /* IP ID */0,
          /* IP Frag */0x40,
          /* TTL */IPPROTO_TCP,
          /* protocol */0,
          /* checksum */htonl (LocalIp),
          htonl (RemoteIp),
          NULL,
          /* payload */0,
          /* payload size */l,
          0
          );
  if (-1 == Tag) {
    return NULL;
  }

  Tag = libnet_build_ethernet (
          RemoteEthaddr,
          LocalEthaddr,
          ETHERTYPE_IP,
          NULL,
          0,
          l,
          0
          );
  if (-1 == Tag) {
    return NULL;
  }

  return l;
}

STATIC
FIELD_T *
TcpUnpackPacket (
  IN UINT8  *Packet,
  IN UINT32 Length
  )
/*++

Routine Description:

  Unpack a TCP packet

Arguments:

  Packet  - The packet should be unpacked
  Length  - The size of the packet

Returns:

  NULL if error occurs
  The pointer to a FIELD_t type's object if successfully

--*/
{
  TCP_HEADER  *Header;
  UINT32      PayloadLen;
  UINT32      OptionsLen;

  if (Length < (sizeof (TCP_HEADER) + sizeof (IP_HEADER) + sizeof (ETH_HEADER))) {
    return NULL;
  }

  Header        = (TCP_HEADER *) (Packet + sizeof (ETH_HEADER) + sizeof (IP_HEADER));

  TcpSrcPort    = ntohs (Header->TcpSrcPort);
  TcpDstPort    = ntohs (Header->TcpDstPort);
  TcpSeq        = ntohs (Header->TcpSeq);
  TcpAck        = ntohs (Header->TcpAck);
  TcpDataOffset = (Header->TcpDataOffset >> 4) & 0x0F;
  TcpCtrlFlag   = Header->TcpCtrlFlag;
  TcpWinSize    = ntohs (Header->TcpWinSize);
  TcpChecksum   = ntohs (Header->TcpChecksum);
  TcpUrgent     = ntohs (Header->TcpUrgent);

  if (NULL != TcpPayload.Payload) {
    free (TcpPayload.Payload);
    TcpPayload.Payload  = NULL;
    TcpPayload.Len      = 0;
  }

  if (NULL != TcpOptions.Payload) {
    free (TcpOptions.Payload);
    TcpOptions.Payload  = NULL;
    TcpOptions.Len      = 0;
  }

  /*
   * data offset is 4 bits width.
   */
  if (TcpDataOffset < 5 || TcpDataOffset > 15) {
    OptionsLen = 0;
  } else {
    OptionsLen = TcpDataOffset * 4 - LIBNET_TCP_H;
  }

  if (OptionsLen) {
    TcpOptions.Payload = malloc (OptionsLen);
    memcpy (
      TcpOptions.Payload,
      Packet + sizeof (TCP_HEADER) + sizeof (IP_HEADER) + sizeof (ETH_HEADER),
      OptionsLen
      );
    TcpOptions.Len = OptionsLen;
  }

  PayloadLen = Length - OptionsLen - sizeof (TCP_HEADER) - sizeof (IP_HEADER) - sizeof (ETH_HEADER);
  if (PayloadLen) {
    TcpPayload.Payload = malloc (PayloadLen);
    memcpy (
      TcpPayload.Payload,
      Packet + OptionsLen + sizeof (TCP_HEADER) + sizeof (IP_HEADER) + sizeof (ETH_HEADER),
      PayloadLen
      );
    TcpPayload.Len = PayloadLen;
  }

  return TcpField;
}
